// FD_Main.cpp
// coded by Christopher Holmes for UMass CVRL, 23APR2000

#include <stdio.h>
#include "FlightData.h"

/////////////////////////////////////////////////////////////////////////////

int main()
{
  char buf[4];
  CFlightData *pFD = new CFlightData();

  if (!pFD)
    {
      printf("Memory Error\n");
      return (0);
    }

  do
    {
      printf("\nFlightData1.0 : Flight Data Manager\n\n");

      printf("Select an option\n\n");

      printf(" 1. Clean a Raw Horita File\n");
      printf(" 2. Clean a Raw Laser File\n");
      printf(" 3. Clean a Raw Watson File\n\n");

      printf(" 4. Read a Raw Horita File\n");
      printf(" 5. Compute a Ground Point\n\n");

      printf(" 6. Compose a Data Sheet\n\n");

      printf(" A. Analysis\n\n");

      printf("Select ('q' to quit): ");
      fgets(buf,4,stdin);
      printf("\n\n");
      pFD->ResetDocument();

      if (buf[0] == 'A' || buf[0] == 'a')
        {
          printf("Analysis options (gnuplot files generated - (y vs. x)):\n\n");

          printf("Meta-File plots (zero values are omitted from plot):\n");
          printf(" A. Ground Altitude Plot (Meters from Sea Level v. Distance)\n");
          printf(" G. Ground Point Plot (Northing v. Easting)\n");
          printf(" P. Plane Plot (Northing v. Easting)\n\n");

          printf("mSec/Byte v. TimeCode analysis (capitals mean un-evaluated):\n");
          printf(" h/H. Horita Plot Analysis\n");
          printf(" l/L. Laser Plot Analysis\n");
          printf(" w/W. Watson Plot Analysis\n\n");

          printf("Select ('q' to quit, 'm' for main menu): ");
          fgets(buf,4,stdin);
          printf("\n\n"); 
        }

      switch (buf[0])
        {
        case 'Q':
        case 'q':
          printf("GoodBye\n");
          delete pFD;
          return (0);
        case '1':
          printf("Enter the Raw Horita File to be Cleaned: ");
          fgets(pFD->m_inputHoritaFile, 255, stdin);
          sscanf(pFD->m_inputHoritaFile,"%s",pFD->m_inputHoritaFile);
          printf("Enter the filename for the Clean Output (no extension): ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Cleaning Raw Horita File...\n");
          if (pFD->CleanRawHorita())
            printf("Raw Horita File has been cleaned\n\n");
          break;
        case '2':
          printf("Enter the Raw Laser File to be Cleaned: ");
          fgets(pFD->m_inputLaserFile, 255, stdin);
          sscanf(pFD->m_inputLaserFile,"%s",pFD->m_inputLaserFile);
          printf("Enter the filename for the Clean Output (no extension): ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Cleaning Raw Laser File...\n");
          if (pFD->CleanRawLaser())
            printf("Raw Laser File has been cleaned\n\n");
          break;
        case '3':
          printf("Enter the Raw Watson File to be Cleaned: ");
          fgets(pFD->m_inputWatsonFile, 255, stdin);
          sscanf(pFD->m_inputWatsonFile,"%s",pFD->m_inputWatsonFile);
          printf("Enter the filename for the Clean Output (no extension): ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Cleaning Raw Watson File...\n");
          if (pFD->CleanRawWatson())
            printf("Raw Watson File has been cleaned\n\n");
          break;
        case '4':
          printf("Enter the Raw Horita File to be Converted: ");
          fgets(pFD->m_inputHoritaFile, 255, stdin);
          sscanf(pFD->m_inputHoritaFile,"%s",pFD->m_inputHoritaFile);
          printf("Enter the filename for the Clean Output (no extension): ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Converting Raw Horita File...\n");
          if (pFD->ReadRawHorita())
            printf("Raw Horita File has been converted for reading\n\n");
          break;
        case '5':
          {
            char buf[32];
            int frame;
            float roll, pitch;
            double n1, n2, e1, e2, laser;

            printf("This procedure will compute a single ground point:\n\n");

            printf("Enter the first UTM Northing: ");
            fgets(buf,31,stdin);
            if (sscanf(buf,"%lf",&n1) != 1) break;
            if (n1 < 0) break;

            printf("Enter the first UTM Easting: ");
            fgets(buf,31,stdin);
            if (sscanf(buf,"%lf",&e1) != 1) break;
            if (e1 < 0) break;

            printf("Enter the second UTM Northing: ");
            fgets(buf,31,stdin);
            if (sscanf(buf,"%lf",&n2) != 1) break;
            if (n2 < 0) break;

            printf("Enter the second UTM Easting: ");
            fgets(buf,31,stdin);
            if (sscanf(buf,"%lf",&e2) != 1) break;
            if (e2 < 0) break;

            printf("Enter the laser value: ");
            fgets(buf,31,stdin);
            if (sscanf(buf,"%lf",&laser) != 1) break;
            if (laser < 0) break;

            printf("Enter the aircraft roll: ");
            fgets(buf,31,stdin);
            if (sscanf(buf,"%f",&roll) != 1) break;
            if (roll <= -90.0 || roll >= 90.0) break;

            printf("Enter the aircraft pitch: ");
            fgets(buf,31,stdin);
            if (sscanf(buf,"%f",&pitch) != 1) break;
            if (pitch <= -90.0 || pitch >= 90.0) break;

            printf("Enter the Horita frame number: ");
            fgets(buf,31,stdin);
            if (sscanf(buf,"%d",&frame) != 1) break;
            if (frame < 0 || frame > 30) break;

            pFD->ComputeGP(n1, e1, n2, e2, laser, roll, pitch, frame);
            break;
          }
        case 'H':
          pFD->m_original = true;
        case 'h':
          printf("Enter the Raw Horita File to be Analyzed: ");
          fgets(pFD->m_inputHoritaFile, 255, stdin);
          sscanf(pFD->m_inputHoritaFile,"%s",pFD->m_inputHoritaFile);
          printf("Enter the filename for the Plot Output (no extension): ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Producing Analysis of Horita File...\n");
          pFD->m_analysis = true;
          if (pFD->CleanRawHorita())
            printf("Analysis complete\n\n");
          break;
        case 'L':
          pFD->m_original = true;
        case 'l':
          printf("Enter the Raw Laser File to be Analyzed: ");
          fgets(pFD->m_inputLaserFile, 255, stdin);
          sscanf(pFD->m_inputLaserFile,"%s",pFD->m_inputLaserFile);
          printf("Enter the filename for the Plot Output (no extension): ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Producing Analysis of Raw Laser File...\n");
          pFD->m_analysis = true;
          if (pFD->CleanRawLaser())
            printf("Analysis complete\n\n");
          break;
        case 'W':
          pFD->m_original = true;
        case 'w':
          printf("Enter the Raw Watson File to be Analyzed: ");
          fgets(pFD->m_inputWatsonFile, 255, stdin);
          sscanf(pFD->m_inputWatsonFile,"%s",pFD->m_inputWatsonFile);
          printf("Enter the filename for the Plot Output (no extension): ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Producing Analysis of Raw Watson File...\n");
          pFD->m_analysis = true;
          if (pFD->CleanRawWatson())
            printf("Analysis complete\n\n");
          break;
        case '6':
	  {
	    bool error = false;
	    char input[24];
	    // Initializations
	    pFD->m_composerFlags[cLASER_FILE] = false;
	    pFD->m_composerFlags[cLASER_AVE] = false;
	    pFD->m_composerFlags[cLASER_ALT] = false;
	    pFD->m_composerFlags[cLASER_GND] = false;
	    pFD->m_laserValue = 0.0;
	    pFD->m_composerFlags[cWATSON_FILE] = false;
	    pFD->m_rollOffset = 0.0;
	    pFD->m_pitchOffset = 0.0;
	    pFD->m_cameraForward = 0.0;
	    pFD->m_cameraSide = 0.0;
	    pFD->m_cameraVertical = 0.0;
	    pFD->m_composerFlags[cLASER_ONLY] = true;
	    pFD->m_composerFlags[cNO_ZEROS] = false;
	    pFD->m_composerFlags[cORIG_HEADING] = false;
	    pFD->m_markLines = false;

	    printf("This procedure will generate a data sheet from\n");
	    printf("   cleaned data files and user input values:\n\n");

	    // LASER
	    printf("Laser Rangefinder options:\n");
	    printf("\t1. input Laser file\n");
	    printf("\t2. enter average Laser value\n");
	    printf("\t3. enter average distance from plane to ground");
	    printf("\t4. enter average ground Altitude\n");
	    printf("Selection: ");
	    fgets(input,3,stdin);
	    switch (input[0])
	      {
	      case '1':
		printf("Enter the clean Laser file: ");
		fgets(pFD->m_inputLaserFile, 255, stdin);
		sscanf(pFD->m_inputLaserFile,"%s",pFD->m_inputLaserFile);
		pFD->m_composerFlags[cLASER_FILE] = true;
		break;
	      case '2':
		do
		  {
		    printf("Enter the average Laser value (feet): ");
		    fgets(input,23,stdin);
		  }
		while (sscanf(input,"%f",&(pFD->m_laserValue)) != 1);
		pFD->m_composerFlags[cLASER_AVE] = true;
		break;
	      case '3':
		do
		  {
		    printf("Enter the average distance from plane to ground (feet): ");
		    fgets(input,23,stdin);
		  }
		while (sscanf(input,"%f",&(pFD->m_laserValue)) != 1);
		pFD->m_composerFlags[cLASER_ALT] = true;
		break;
	      case '4':
		do
		  {
		    printf("Enter the average ground Altitude (meters from Sea Level): ");
		    fgets(input,23,stdin);
		  }
		while (sscanf(input,"%f",&(pFD->m_laserValue)) != 1);
		pFD->m_composerFlags[cLASER_GND] = true;
		break;
	      default:
		printf("Watch those fingers, Sloppy Joe");
		error = true;
	      }
	    if (error) break;

	    // WATSON
	    printf("\nDo you have a Watson file? (y/n): ");
	    fgets(input,4,stdin);
	    if (input[0] == 'y' || input[0] == 'Y')
	      {
		pFD->m_composerFlags[cWATSON_FILE] = true;
		printf("Enter the clean Watson file: ");
		fgets(pFD->m_inputWatsonFile, 255, stdin);
		sscanf(pFD->m_inputWatsonFile,"%s",pFD->m_inputWatsonFile);
		printf("Enter any Roll offset: ");
		fgets(input,23,stdin);
		if (sscanf(input,"%f",&(pFD->m_rollOffset)) != 1) break;
		printf("Enter any Pitch offset: ");
		fgets(input,23,stdin);
		if (sscanf(input,"%f",&(pFD->m_pitchOffset)) != 1) break;
	      }

	    // HORITA
	    printf("Enter the clean Horita File: ");
	    fgets(pFD->m_inputHoritaFile, 255, stdin);
	    sscanf(pFD->m_inputHoritaFile,"%s",pFD->m_inputHoritaFile);

	    printf("The following offsets relate the camera to the GPS antenna...\n");
	    printf("Enter the camera FORWARD offset (+ ahead): ");
	    fgets(input,23,stdin);
	    if (sscanf(input,"%f",&(pFD->m_cameraForward)) != 1) break;

	    printf("Enter the camera SIDE offset (+ right): ");
	    fgets(input,23,stdin);
	    if (sscanf(input,"%f",&(pFD->m_cameraSide)) != 1) break;

	    printf("Enter the camera VERTICAL offset (+ above): ");
	    fgets(input,23,stdin);
	    if (sscanf(input,"%f",&(pFD->m_cameraVertical)) != 1) break;

	    // OUTPUT
	    printf("Enter the output file path and prefix name: ");
	    fgets(pFD->m_outputFileName, 255, stdin);
	    sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

	    printf("Do you want a (s)ingle output file, or (m)ultiple output files? ");
	    fgets(input,4,stdin);
	    if (input[0] == 'm' || input[0] == 'M')
	      {
		pFD->m_composerFlags[cOUTPUT_SINGLE] = false;
	      }
	    else
	      {
		pFD->m_composerFlags[cOUTPUT_SINGLE] = true;
	      }

	    printf("\nOutput format Options:\n");
	    printf("Sorting in the actual Watson and GPS readings creates new zero\n");
	    printf("Laser values, which can interfere with future Laser processing.\n");
	    printf("Do you want to keep these readings in? (y/n): ");
	    fgets(input,23,stdin);
	    if (input[0] == 'y' || input[0] == 'Y')
	      {
		pFD->m_composerFlags[cLASER_ONLY] = false;
	      }
	    printf("Zeros in the laser or GPS readings will translate to zeros\n");
	    printf("in the computed Ground values.\n");
	    printf("Do you want to remove them? (y/n): ");
	    fgets(input,23,stdin);
	    if (input[0] == 'y' || input[0] == 'Y')
	      {
		pFD->m_composerFlags[cNO_ZEROS] = true;
	      }
	    printf("By default the heading is corrected using the GPS coordinates.\n");
	    printf("Do you want the heading left alone? (y/n): ");
            fgets(input,23,stdin);
	    if (input[0] == 'y' || input[0] == 'Y')
	      {
		pFD->m_composerFlags[cORIG_HEADING] = true;
	      }
	    printf("Each row of data starts with a timestamp, which\n");
	    printf("marks the occurrence of an instrument reading.\n");
	    printf("Do you want the rows labeled by instrument? (y/n) :");
            fgets(input,23,stdin);
	    if (input[0] == 'y' || input[0] == 'Y')
	      {
		pFD->m_markLines = true;
	      }

	    printf("\nRunning Composer...\n");
	    pFD->Composer();
            printf("Composer Done\n");
            break;
	  }
        case 'm':
        case 'M':
          break;
        case 'A':
        case 'a':
          pFD->pType = PLOT_ALT;
          printf("Enter the name of the meta-file from which to plot\n");
          printf(" the ground altitude: ");
          fgets(pFD->m_inputHoritaFile, 255, stdin);
          sscanf(pFD->m_inputHoritaFile,"%s",pFD->m_inputHoritaFile);
          printf("Enter the output plot filename: ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Generating Plot File from %s to %s ...\n",pFD->m_inputHoritaFile,
                 pFD->m_outputFileName);
          pFD->GenPlot();
          printf("Plotting Done\n");
          break;
        case 'G':
        case 'g':
          pFD->pType = PLOT_GROUND;
          printf("Enter the name of the meta-file from which to plot\n");
          printf(" the ground coordinates: ");
          fgets(pFD->m_inputHoritaFile, 255, stdin);
          sscanf(pFD->m_inputHoritaFile,"%s",pFD->m_inputHoritaFile);
          printf("Enter the output plot filename: ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Generating Plot File...\n");
          pFD->GenPlot();
          printf("Plotting Done\n");
          break;
        case 'P':
        case 'p':
          pFD->pType = PLOT_PLANE;
          printf("Enter the name of the meta-file from which to plot\n");
          printf(" the airplane's coordinates: ");
          fgets(pFD->m_inputHoritaFile, 255, stdin);
          sscanf(pFD->m_inputHoritaFile,"%s",pFD->m_inputHoritaFile);
          printf("Enter the output plot filename: ");
          fgets(pFD->m_outputFileName, 255, stdin);
          sscanf(pFD->m_outputFileName,"%s",pFD->m_outputFileName);

          printf("Generating Plot File...\n");
          pFD->GenPlot();
          printf("Plotting Done\n");
          break;
        default:
          printf("Not a valid option, try again\n\n");
        }
    }
  while (1);

  return (0);
}
