// FlightData.cpp : implementation of the CFlightData class
//

#include <stdio.h>
#include <string.h>
#include "FlightData.h"

/////////////////////////////////////////////////////////////////////////////
// CFlightData

/////////////////////////////////////////////////////////////////////////////
// CFlightData construction/destruction

CFlightData::CFlightData()
{
  // One time construction code
  m_listHead = 0;
  fpLAS = 0;
  fpWAT = 0;
  fpHOR = 0;
  fpOUT = 0;
  m_masterList = 0;

  // one time composer variable setting
  m_composerFlags[cLASER_FILE] = true;
  m_composerFlags[cLASER_AVE] = false;
  m_composerFlags[cLASER_ALT] = false;
  m_composerFlags[cLASER_GND] = false;
  m_composerFlags[cWATSON_FILE] = true;
  m_composerFlags[cOUTPUT_SINGLE] = true;
  m_composerFlags[cLASER_ONLY] = false;
  m_composerFlags[cNO_ZEROS] = false;
  m_composerFlags[cORIG_HEADING] = false;
  m_cameraForward = 0.0;
  m_cameraSide = 0.0;
  m_cameraVertical = 0.0;
  m_rollOffset = 0.0;
  m_pitchOffset = 0.0;
  m_laserValue = 0.0;
}

CFlightData::~CFlightData()
{
  ClearList();
  CloseFiles();
}

/////////////////////////////////////////////////////////////////////////////
// CFlightDataDoc commands

void CFlightData::ResetDocument()
{
  ClearList();
  CloseFiles();
  m_currentID = NONE;
  m_analysis = m_original = false;
  m_inputLaserFile[0] = '\0';
  m_inputHoritaFile[0] = '\0';
  m_inputWatsonFile[0] = '\0';
  m_outputFileName[0] = '\0';

  return;
}

void CFlightData::ClearList()
{
  ENTRY *temp;
  LAB *tmp;

  while (m_listHead)
    {
      temp = m_listHead;
      m_listHead = m_listHead->next;
      delete temp;
    }

  while (m_masterList)
    {
      tmp = m_masterList;
      m_masterList = m_masterList->next;
      delete tmp;
    }

  return;
}

ENTRY * CFlightData::GetEntry()
{
  ENTRY *temp = new ENTRY;
  if (!temp) return NULL;

  memset(temp,'\0',sizeof(ENTRY));
  return (temp);
}

LAB * CFlightData::GetLab()
{
  LAB *temp;

  temp = new LAB;
  if (!temp) return NULL;

  memset(temp,'\0',sizeof(LAB));
  return (temp);
}

void CFlightData::Out2User(char * s)
{
  printf("%s\n",s);
  return;
}

void CFlightData::CloseFiles()
{
  if (fpLAS) fclose(fpLAS);
  fpLAS = 0;
  if (fpHOR) fclose(fpHOR);
  fpHOR = 0;
  if (fpWAT) fclose(fpWAT);
  fpWAT = 0;
  if (fpOUT) fclose(fpOUT);
  fpOUT = 0;
}

// This procedure computes the milliseconds/byte for each pair of consecutive timestamps,
// associating the rate with the first timestamp. This procedure also monitors the gap
// between each timestamp, and if there's an error, it will return a pointer to the
// timestamp prior to the error. Otherwise, if all go well, this procedure returns a zero.
ENTRY * CFlightData::ComputeRate()
{
  ENTRY *thisTime, *nextTime;
  int sum, totalSum = 0;
  double tdiff, totalTime = 0.0;

  // calculate milliseconds/byte for each time (from that time to the next time)
  for (thisTime = m_listHead; thisTime;)
    {
      sum = strlen(thisTime->data) - thisTime->offset;

      // CHECK FOR BREAKS IN THE HORITA BY LOOKING AT THE GPS TIMECODE
      if (m_currentID == HORITA)
        {
          ENTRY *tracker = thisTime;
          for (nextTime = thisTime->next; nextTime && nextTime->labtime == 0.0; nextTime = nextTime->next)
            {
              sum += strlen(nextTime->data);
              if (strlen(tracker->data) < 14 || strlen(nextTime->data) < 14)
                {
                  nextTime = nextTime->next;
                  if (nextTime) return (thisTime);
                  break; // this is the end
                }
              int hr1 = (tracker->data[12] - 48) * 10 + (tracker->data[13] - 48);		// hour
              int hr2 = (nextTime->data[12] - 48) * 10 + (nextTime->data[13] - 48);	// hour
              int mn1 = (tracker->data[9] - 48) * 10 + (tracker->data[10] - 48);		// minute
              int mn2 = (nextTime->data[9] - 48) * 10 + (nextTime->data[10] - 48);	// minute
              int sc1 = (tracker->data[6] - 48) * 10 + (tracker->data[7] - 48);		// second
              int sc2 = (nextTime->data[6] - 48) * 10 + (nextTime->data[7] - 48);		// second
              int fr1 = (tracker->data[3] - 48) * 10 + (tracker->data[4] - 48);		// frame
              int fr2 = (nextTime->data[3] - 48) * 10 + (nextTime->data[4] - 48);		// frame
              if (hr1 == hr2) // hour
                {
                  if (mn1 == mn2) // minute
                    {
                      if (sc1 == sc2) // second
                        {
                          // hour, minute, and second are the same
                          if (fr2 < fr1 || fr2 > fr1+FRAME_TOL) // if frames are not within FRAME_TOL of each other
                            return (thisTime);
                        }
                      else if (sc1+1 != sc2 || fr1 != 29 || fr2 > 4) // second was not advanced properly
                        {
                          return (thisTime);
                        }
                    }
                  // minute was not advanced properly
                  else if (mn1+1 != mn2 || sc1 != 59 || sc2 != 0 || fr1 != 29 && fr2 > 4)
                    return (thisTime);
                }
				// hour was not advanced properly
              else if (hr1+1 != hr2 || mn1 != 59 || mn2 != 0 || sc1 != 59 || sc2 != 0 || fr1 != 29 && fr2 > 4)
                return (thisTime);

              tracker = nextTime;
            }
        }
      else
        {
          for (nextTime = thisTime->next; nextTime && nextTime->labtime == 0.0; nextTime = nextTime->next)
            sum += strlen(nextTime->data);
        }

      if (!nextTime) break;

      sum += nextTime->offset;
      tdiff = nextTime->labtime - thisTime->labtime;

      if (tdiff < 0) return (thisTime);

      // CHECK FOR BREAKS IN THE LASER AND WATSON BY LOOKING FOR LARGE GAPS IN THE TIMESTAMP SEQUENCE
      switch(m_currentID)
        {
        case LASER:
          if (tdiff > LASER_TOLERANCE) return (thisTime);
          break;
        case WATSON:
          if (tdiff > WATSON_TOLERANCE) return (thisTime);
          break;
        }

      //	if everything seems consistant to this point, compute the rate and sum for the running average
      if (sum)
        thisTime->mSecPerByte = tdiff / (double) sum;

      totalTime += tdiff;
      totalSum += sum;

      thisTime = nextTime;
    }

  // if everything checked out OK in this queue, update the running average
  if (averageSum < 10000000.0 && averageTime < 10000000.0)
    {
      averageSum += (double) totalSum;
      averageTime += totalTime;
    }
  return (0);
}

// This procedure monitors the ComputeRate function, taking action if an error in the timestamp
// sequence is found. If there is an error, this function returns a zero, otherwise it returns
// the last timestamp in the list.
ENTRY * CFlightData::PrevalQueue()
{
  ENTRY *thisTime, *nextTime;
  static bool Toggle = true;
  short strikes, x;
  char *oldFile;

  for (strikes = 3; strikes; strikes--)
    {
      // compute the milliseconds-per-byte for each timestamp
      thisTime = ComputeRate();

      if (m_original) return (0);
      // if no pointer was returned, there was no problem - break out and return
      if (!thisTime) break;

      // This will complete processing on any of the list prior to the time before
      // thisTime, and output to the current file. Toggle prevents an infinite loop.
      if (Toggle)
        {
          Toggle = false;
          return (thisTime);
        }
      Toggle = true;

      // delete ERROR_BUFFER number of timestamps before starting the new file
      for (x = 0, nextTime = thisTime; nextTime; nextTime = nextTime->next)
        {
          if (nextTime->labtime != 0.0)
            {
              x++;
              while (m_listHead != nextTime)
                {
                  thisTime = m_listHead;
                  m_listHead = m_listHead->next;
                  delete thisTime;
                }
              thisTime = m_listHead;
            }
          if (x >= ERROR_BUFFER)
            {
              x = 0;
              break;
            }
        }

      if (!nextTime) Out2User("Warning, Queue Size too small to allow proper Error Buffering");

      // if analyzing, just put a break in the data, else separate the files
      if (m_analysis)
        {
          fprintf(fpOUT,"\n");
        }
      else
        {
          // close the current output file, and create/open a new output file
          fclose(fpOUT);
          fpOUT = 0;

          sprintf(m_newFile,"%s%d.txt",m_outputFileName, m_fileNumber-1);
          oldFile = strdup(m_newFile);
          if (oldFile)
            {
              sprintf(m_newFile,"%s%d_%1.1lf-%1.1lf.txt",m_outputFileName, m_fileNumber-1, m_startLab, m_endLab);
              rename(oldFile, m_newFile);
              delete[] oldFile;
            }
          m_startLab = m_listHead->labtime;

          sprintf(m_newFile,"%s%d.txt",m_outputFileName, m_fileNumber++);
          fpOUT = fopen(m_newFile,"w");
          if (!fpOUT)
            {
              if (m_currentID == LASER)
                Out2User("Cannot Open a new Clean Laser File");
              if (m_currentID == WATSON)
                Out2User("Cannot Open a new Clean Watson File");
              if (m_currentID == HORITA)
                Out2User("Cannot Open a new Clean Horita File");
              CloseFiles();
              return (0);
            }
        }
    }
  if (!strikes)
    {
      Out2User("Three Strikes in a Queue - Horrible Data!");
      return (0);
    }
  thisTime = m_listHead;
  for (nextTime = m_listHead; nextTime->next; nextTime = nextTime->next)
    {
      if (nextTime->labtime != 0.0) thisTime = nextTime;
    }
  return (thisTime);
}

// this procedure will write out and delete all info in the queue
// up to the 'lastTime' in the queue.
void CFlightData::Out2File(ENTRY * lastTime)
{
  ENTRY *temp = 0;
  int count;
  double rate, time = 0.0;

  if (!m_listHead || !fpOUT) return;

  if (m_analysis)
    {
      while (m_listHead != lastTime)
        {
          if (m_listHead->labtime != 0.0)
            fprintf(fpOUT,"%1.2lf\t%1.2lf\n",m_listHead->labtime, m_listHead->mSecPerByte);
          temp = m_listHead;
          m_listHead = temp->next;
          delete temp;
        }
      return;
    }

  do
    {
      if (temp)
        {
          m_listHead = m_listHead->next;
          delete temp;
        }

      //output time and data
      if (m_listHead && m_listHead->data[0] && time)
        {
          m_endLab = time + (rate * (double)count);
          switch (m_currentID)
            {
            case LASER:
              {
                int scatter;
                double laser;
                if (sscanf(m_listHead->data,"%lff,%d",&laser, &scatter) != 2)
                  {
                    sprintf(m_newFile,"Tossing out bad data from around time = %1.1lf:%s:",time, m_listHead->data);
                    Out2User(m_newFile);
                  }
                else
                  {
                    fprintf(fpOUT,"%1.2lf\t%7.1lf\t%3d\n", m_endLab, laser, scatter);
                  }
              }
              break;
            case WATSON:
              if (m_listHead->data[0] != 'I' && m_listHead->data[0] != 'i')
                {
                  sprintf(m_newFile,"Tossing out bad data from around time = %1.1lf:%s:",time, m_listHead->data);
                  Out2User(m_newFile);
                  break;
                }
            case HORITA:
              fprintf(fpOUT,"%1.2lf\t%s\n", m_endLab, m_listHead->data);
              break;
            case HORITA_RR:
              if (m_listHead != lastTime)
                {
                  int i, frame, sec, min, hour, u1, u2, u3, u4;
                  i = sscanf(m_listHead->data,"%*s %d %d %d %d %*s %d %d %d %d", &frame, &sec, &min, &hour, &u4, &u3, &u2, &u1);
                  if (i != 8)
                    {
                      i = sscanf(m_listHead->data,"%*s %d %d %d %d", &frame, &sec, &min, &hour);
                      if (i != 4)
                        fprintf(fpOUT,"Scanning error:%s:\n",m_listHead->data);
                      else
                        {
                          fprintf(fpOUT,"%9.1lf\t%d:%d:%d:%d\tUser Data = %s\n", m_listHead->labtime, hour, min, sec, frame,
                                  &(m_listHead->data[14]));
                        }
                    }
                  else
                    {
                      fprintf(fpOUT,"%9.1lf\t%d:%d:%d:%d\t", m_listHead->labtime, hour, min, sec, frame);
                      ProcessUserData(frame, u1, u2, u3, u4);
                    }
                }
              break;
            }

          count += strlen(m_listHead->data);
        }

      if (m_listHead && m_listHead->labtime != 0.0)
        {
          time = m_listHead->labtime;
          rate = m_listHead->mSecPerByte;
          count = strlen(m_listHead->data) - m_listHead->offset;
        }
      temp = m_listHead;
    }
  while (m_listHead != lastTime);

  return;
}

ENTRY * CFlightData::RemoveSpikes(ENTRY * theEnd)
{
  ENTRY *thisTime, *nextTime, *preNextTime;
  int sum;
  double tdiff, average = 0.0, newaverage = 0.0, spikeTolerance = 0.1;

  if (averageSum) average = averageTime / averageSum;

  if (averageSum < 5000)
    {
      sprintf(m_newFile,"Average = %lf\n",average);
      Out2User(m_newFile);
    }

  if (m_currentID == LASER) spikeTolerance = 0.03;
  if (m_currentID == WATSON) spikeTolerance = 0.05;
  if (m_currentID == HORITA) spikeTolerance = 0.02;
  // seek out and remove all spikes that range beyond the average +/- spikeTolerance,
  // and recalculate the previous time's mSec/Byte.
  for (thisTime = m_listHead; thisTime && thisTime != theEnd;)
    {
      sum = strlen(thisTime->data) - thisTime->offset;

      // get next time
      preNextTime = thisTime;
      for (nextTime = thisTime->next; nextTime && nextTime->labtime == 0.0; nextTime = nextTime->next)
        {
          preNextTime = nextTime;
          sum += strlen(nextTime->data);
        }

      if (!nextTime || nextTime == theEnd) break;

      // seek out the spike: two consecutive milliseconds that exceed the tolerance at opposite ends
      if ((thisTime->mSecPerByte < average-spikeTolerance && nextTime->mSecPerByte > average+spikeTolerance) ||
          (thisTime->mSecPerByte > average+spikeTolerance && nextTime->mSecPerByte < average-spikeTolerance))
        {
          // continue to next time, leaving the preNextTime
          sum += strlen(nextTime->data);

          for (nextTime = nextTime->next; nextTime && nextTime->labtime == 0.0; nextTime = nextTime->next)
            {
              sum += strlen(nextTime->data);
            }

          if (!nextTime) break;

          sum += nextTime->offset;
          if (sum)
            {
              tdiff = nextTime->labtime - thisTime->labtime;
              newaverage = tdiff / (double) sum;

              // since a spike was found this newaverage will always be better
              thisTime->mSecPerByte = newaverage;
              // remove old nextTime
              if (preNextTime->next->data[0])
                {
                  preNextTime = preNextTime->next;
                  preNextTime->labtime = 0.0;
                  preNextTime->mSecPerByte = 0.0;
                  preNextTime->offset = 0;
                }
              else
                {
                  thisTime = nextTime;
                  nextTime = preNextTime->next;
                  preNextTime->next = nextTime->next;
                  delete nextTime;
                  continue;
                }
            }
        }
      thisTime = nextTime;
    }
  return (thisTime);
}

ENTRY * CFlightData::RemovePeaks(ENTRY * theEnd)
{
  ENTRY *thisTime, *nextTime, *tracker;
  int sum;
  double average = 0.0, newaverage = 0.0, peakTolerance = 1.0;

  if (averageSum) average = averageTime / averageSum;

  if (m_currentID == LASER) peakTolerance = 0.09;
  if (m_currentID == WATSON) peakTolerance = 0.05;
  if (m_currentID == HORITA) peakTolerance = 0.02;
  // seek out and remove all times who's mSec/Byte is beyond the average +/- peakTolerance
  // and recalculate the previous time's mSec/Byte
  for (thisTime = m_listHead; thisTime && thisTime != theEnd;)
    {
      sum = strlen(thisTime->data) - thisTime->offset;

      // get next time
      tracker = thisTime;
      for (nextTime = thisTime->next; nextTime && nextTime->labtime == 0.0; nextTime = nextTime->next)
        {
          tracker = nextTime;
          sum += strlen(nextTime->data);
        }

      if (!nextTime || nextTime == theEnd) break;

      // remove peaks
      if (nextTime->mSecPerByte > average+peakTolerance || thisTime->mSecPerByte < average-peakTolerance)
        {
          sum += strlen(nextTime->data);

          // remove
          if (nextTime->data[0])
            {
              nextTime->labtime = 0.0;
              nextTime->mSecPerByte = 0.0;
              nextTime->offset = 0;
              tracker = nextTime;
            }
          else
            {
              tracker->next = nextTime->next;
              delete nextTime;
            }

          // recalculate
          for (nextTime = tracker->next; nextTime && nextTime->labtime == 0.0; nextTime = nextTime->next)
            {
              sum += strlen(nextTime->data);
            }

          if (!nextTime) break;

          sum += nextTime->offset;
          if (sum)
            thisTime->mSecPerByte = (nextTime->labtime - thisTime->labtime) / (double) sum;
        }
      thisTime = nextTime;
    }
  // return a pointer to the ENTRY who's rate was not computed
  // this way, the printout function will know where to stop
  return (thisTime);
}

bool CFlightData::CleanRawLaser()
{
  m_currentID = LASER;
  return CleanRawFile();
}

int CFlightData::LoadRawLaser()
{
  bool badData, needNewEntry, partialData = false;
  int index, num;
  char buf[64];
  ENTRY *here, *tail, *newEntry = 0;

  if (!m_listHead || !fpLAS) return (0);

  // count the number of existing entries
  for (tail = m_listHead, num = 1; tail->next; tail = tail->next, num++);
  here = tail;

  // LOAD THE QUEUE
  needNewEntry = true;
  while (num < MAXSIZE || partialData == true)
    {
      // ensure we have the memory before scanning the file for data
      if (needNewEntry)
        {
          newEntry = GetEntry();
          if (!newEntry) return num;
          needNewEntry = false;
        }
      // scan a line from the input file (if end of file then break)
      if (fgets(buf,63,fpLAS) == 0) break;
      badData = false;

      // character check to avoid header data, and also to remove the return character
      for (index = 0; buf[index]; index++)
        {
          if (buf[index] == '\n' || buf[index] == '\r')
            {
              buf[index] = '\0';
              break;
            }
          // if it's not laser data, it's bad data
          if (!(buf[index] <= '9' && buf[index] >= '0') && buf[index] != 'T'
              && buf[index] != 'f' && buf[index] != '.' && buf[index] != ','
              && buf[index] != ' ')
            {
              sprintf(m_newFile,"Garbage found:%d: %s\n",buf[index],buf);
              Out2User(m_newFile);
              badData = true;
              partialData = false;
              break;
            }
        }
      if (badData) continue;

      // look for timestamp 'T' key
      for (index = 0; buf[index] && buf[index] != 'T'; index++);

      // if no 'T', it's just a line of laser data
      if (!buf[index])
        {
          if (partialData) // add to existing data entry
            {
              strncpy(&(here->data[here->offset]), buf, 63 - here->offset);
              partialData = false;
              continue;
            }

          // else
          if (index == 0) continue;
          here = newEntry;
          newEntry = 0;
          needNewEntry = true;
          here->offset = strlen(buf);
          strncpy(here->data, buf, 63);
        }

      else // a 'T' was found
        {
          here = newEntry;
          newEntry = 0;
          needNewEntry = true;

          // extract the time
          sscanf(&(buf[index]),"T%lf",&(here->labtime));

          here->offset = index;
          // if the 'T' is not the first character, there's other data before it
          if (index != 0)
            {
              buf[index] = '\0';
              strncpy(here->data, buf, 63);
              partialData = true;
            }
        }

      // add to queue
      tail->next = here;
      tail = here;
      num++;
    }

  if (newEntry) delete newEntry;
  return num;
}

bool CFlightData::CleanRawWatson()
{
  m_currentID = WATSON;
  return CleanRawFile();
}

int CFlightData::LoadRawWatson()
{
  bool badData, needNewEntry, partialData = false;
  int index, num;
  char buf[64];
  ENTRY *here, *tail, *newEntry = 0;

  if (!m_listHead || !fpWAT) return (0);

  // count the number of existing entries
  for (tail = m_listHead, num = 1; tail->next; tail = tail->next, num++);
  here = tail;

  // LOAD THE QUEUE
  needNewEntry = true;
  while (num < MAXSIZE || partialData == true)
    {
      // ensure we have the memory before scanning the file for data
      if (needNewEntry)
        {
          newEntry = GetEntry();
          if (!newEntry) return num;
          needNewEntry = false;
        }
      // scan a line from the input file (if end of file then break)
      if (fgets(buf,63,fpWAT) == 0) break;
      badData = false;

      // character check to avoid header data, and also to remove the return character
      for (index = 0; buf[index]; index++)
        {
          if (buf[index] == '\n' || buf[index] == '\r')
            {
              buf[index] = '\0';
              break;
            }
          // if it's not watson data, it's bad data
          if (!(buf[index] <= '9' && buf[index] >= '0') && buf[index] != 'T'
              && buf[index] != '-' && buf[index] != '.' && buf[index] != '+'
              && buf[index] != ' ' && buf[index] != 'I' && buf[index] != 'i')
            {
              sprintf(m_newFile,"Garbage found:%d: %s\n",buf[index],buf);
              Out2User(m_newFile);
              badData = true;
              partialData = false;
              break;
            }
        }
      if (badData) continue;

      // look for timestamp 'T' key
      for (index = 0; buf[index] && buf[index] != 'T'; index++);

      // if no 'T', it's just a line of watson data
      if (!buf[index])
        {
          if (partialData) // add to existing data entry
            {
              strncpy(&(here->data[here->offset]), buf, 63 - here->offset);
              partialData = false;
              continue;
            }

          // else
          if (index == 0) continue;
          here = newEntry;
          newEntry = 0;
          needNewEntry = true;
          here->offset = strlen(buf);
          strncpy(here->data, buf, 63);

          if ((buf[0] == 'I' || buf[0] == 'i') && strlen(here->data) < 45) partialData = true;
        }
      
      else // a 'T' was found
        {
          if (!partialData)
            {
              here = newEntry;
              newEntry = 0;
              needNewEntry = true;
            }

          // extract the time
          sscanf(&(buf[index]),"T%lf",&(here->labtime));
          if (partialData) continue;
        }

      // add to queue
      tail->next = here;
      tail = here;
      num++;
    }

  if (newEntry) delete newEntry;
  return num;
}

bool CFlightData::CleanRawHorita()
{
  m_currentID = HORITA;
  return CleanRawFile();
}

int CFlightData::LoadRawHorita()
{
  bool Cvalid = false, Dvalid = false, Evalid = false, check = false;
  int i = 0, num;
  static int index = 0;
  char buf[16];
  unsigned char a, b, c, d, e;
  ENTRY *here, *newEntry = 0;

  if (!m_listHead || !fpHOR) return (0);

  // count the number of existing entries
  for (num = 1, here = m_listHead; here->next; here = here->next, num++);

  // a loop to load the queue
  while (num < MAXSIZE || check || Cvalid)
    {
      // ensure a structure is available from memory
      if (!newEntry)
        {
          newEntry = GetEntry();
          if (!newEntry) return num; // what if check=true || CValid = true ????
        }

      // get character from file (or from previous scan)
      if (Cvalid)
        {
          b = c;
          if (Dvalid) c = d;
          else Cvalid = false;
          if (Evalid) d = e;
          else Dvalid = false;
          Evalid = false;
        }
      else
        {
          if(!fread(&b,1,1,fpHOR)) break;
        }

      if (check) // potential timestamp
        {
          check = false;
          // check if timestamp has been found
          if (a == 'T' && b == ' ')
            {
              // scan and test next 3 values to ensure we have a timestamp
              if (!Cvalid)
                {
                  if (!fread(&c,1,1,fpHOR)) break;
                  Cvalid = true;
                }
              if (!Dvalid)
                {
                  if (!fread(&d,1,1,fpHOR)) break;
                  Dvalid = true;
                }
              if (!Evalid)
                {
                  if (!fread(&e,1,1,fpHOR)) break;
                  Evalid = true;
                }

              // if they are numbers then we have a timestamp
              if (c >= '0' && c <= '9' && d >= '0' && d <= '9' && e >= '0' && e <= '9')
                {
                  buf[0] = c;
                  buf[1] = d;
                  buf[2] = e;
                  if (!fgets(&(buf[3]),12,fpHOR)) break;
                  sscanf(buf,"%lf",&(here->labtime)); // because we know first three values ARE numbers, there should be no error
                  here->offset = index;
                  Cvalid = Dvalid = Evalid = false;
                  a = b = 0;
                  continue;
                }
            }

          // if timestamp check failed, process 'a' here, let 'b' fall through
          if (a) // here a == 'T'
            {
              // process horita data
              here = ProcessHorita(here, &newEntry, &index, &num, a);
            }
        }

      if (b == 'T') // potential timestamp: save this character
        {
          a = b;
          check = true;
          continue;
        }

      // process horita data
      here = ProcessHorita(here, &newEntry, &index, &num, b);
    }
  return num;
}

ENTRY * CFlightData::ProcessHorita(ENTRY * here, ENTRY ** newEntry, int * index, int * num, unsigned char c)
{
  // horita data processing
  int hi = ((int) c) / 16;
  int lo = ((int) c) % 16;

  if (hi == 15 && lo != 15) // HI == F && LO != F
    {
      // create a new structure (if needed);

      // if here->data contains nothing, garbage was
      // found and the data was dumped: just reload here
      if (here->data[0])
        {
          here->next = *newEntry;
          *newEntry = 0;
          here = here->next;
          (*num)++;
        }
      *index = 0;
    }

  if (*index > 40)
    {
      // garbage, the index should not go beyond 30
      // dump it
      if (here->data)
        {
          sprintf(m_newFile,"\nGarbage found:%1.2lf: %s",here->labtime, here->data);
          Out2User(m_newFile);
          here->data[0] = '\0';
        }
      else
        {
          sprintf(m_newFile,"%s %c%c ",m_newFile, Num2HexChar(hi), Num2HexChar(lo));
          Out2User(m_newFile);
        }
    }
  else
    {
      here->data[(*index)++] = Num2HexChar(hi);
      here->data[(*index)++] = Num2HexChar(lo);
      here->data[(*index)++] = ' ';
      here->data[*index] = '\0';
    }
  return (here);
}

char CFlightData::Num2HexChar(int x)
{
  if (x > 15) return (0);
  if (x < 10) return (x+48);
  else return (x+55);
}

bool CFlightData::CleanRawFile()
{
  FILE *fp = 0;
  bool Tfound;
  char line[64], *oldFile;
  unsigned char c, d;
  int i, entryCount;
  ENTRY *here, *gopher, *temp;

  // initialize
  ClearList();
  averageSum = averageTime = 0.0;
  m_fileNumber = 0;

  /////// OPEN BOTH INPUT AND OUTPUT FILES ///////
  switch (m_currentID)
    {
    case LASER:
      fpLAS = fopen(m_inputLaserFile,"r");
      if (!fpLAS)
        {
          Out2User("Cannot Open Raw Laser File");
          return false;
        }
      fp = fpLAS;
      break;
    case WATSON:
      fpWAT = fopen(m_inputWatsonFile,"r");
      if (!fpWAT)
        {
          Out2User("Cannot Open Raw Watson File");
          return false;
        }
      fp = fpWAT;
      break;
    case HORITA:
    case HORITA_RR:
      fpHOR = fopen(m_inputHoritaFile,"r");
      if (!fpHOR)
        {
          Out2User("Cannot Open Raw Horita File");
          return false;
        }
      fp = fpHOR;
      break;
    default:
      Out2User("Internal Error!");
      return false;
    }

  for (i = strlen(m_outputFileName); i >= 0; i--)
    {
      if (m_outputFileName[i] == '.')
        {
          m_outputFileName[i] = '\0';
          break;
        }
      if (m_outputFileName[i] == '\\' || m_outputFileName[i] == '/') break;
    }
  sprintf(m_newFile,"%s%d.txt",m_outputFileName, m_fileNumber++);
  fpOUT = fopen(m_newFile,"w");
  if (!fpOUT)
    {
      Out2User("Cannot Open Clean File");
      CloseFiles();
      return false;
    }

  // CREATE THE FIRST ENTRY AND GET THE FIRST TIMESTAMP
  m_listHead = GetEntry();
  if (!m_listHead)
    {
      Out2User("Memory Error creating first entry");
      CloseFiles();
      return false;
    }

  Tfound = false;
  d = 0;
  while ( fread(&c,1,1,fp) != 0 )
    {
      if (m_currentID == WATSON) // scan for valid Watson data characters
        {
          if (c == '\n' || c == '\r') continue;
          // if it's not watson data, it's bad data
          if (!(c <= '9' && c >= '0') && c != 'T' && c != '-' && c != '.' && c != '+' && c != ' ' && c != 'I' && c != 'i')
            {
              fgets(line, 63, fp);
              sprintf(m_newFile,"Garbage found:%d: %s\n", c, line);
              Out2User(m_newFile);
              continue;
            }
        }
      else if (m_currentID == LASER)
        {
          if (c == '\n' || c == '\r') continue;
          // if it's not laser data, it's bad data
          if (!(c <= '9' && c >= '0') && c != 'T' && c != 'f' && c != '.' && c != ',' && c != ' ')
            {
              fgets(line, 63, fp);
              sprintf(m_newFile,"Garbage found:%d: %s\n", c, line);
              Out2User(m_newFile);
              continue;
            }
        }

      if (d == 0 && c != 'T') continue;
      if (d == 'T')
        {
          if (c == ' ')
            {
              Tfound = true;
              break;
            }
          d = 0;
        }
      if (c == 'T')
        {
          d = c;
        }
    }

  if (!Tfound)
    {
      Out2User("Could not find a timestamp in the raw file");
      CloseFiles();
      return false;
    }

  if (!fgets(line,63,fp)) Tfound = false;
  if (Tfound && sscanf(line,"%lf",&(m_listHead->labtime)) != 1) Tfound = false;
  if (!Tfound)
    {
      Out2User("Error scanning first time from raw File");
      CloseFiles();
      return false;
    }
  // record first labtime
  m_startLab = m_listHead->labtime;

  do
    {
      if (m_currentID == LASER) entryCount = LoadRawLaser();
      else if (m_currentID == WATSON) entryCount = LoadRawWatson();
      else entryCount = LoadRawHorita();

      if (m_currentID == HORITA_RR)
        {
          temp = 0;
        }
      else if (m_original)
        {
          temp = 0;
          ComputeRate();
        }
      else
        {
          temp = PrevalQueue();
          if (!temp)
            {
              Out2User("Cleaning Stopped due to Error in Raw File");
              return false;
            }

          RemoveSpikes(temp);
          RemoveSpikes(temp);
          RemovePeaks(temp);
          temp = RemovePeaks(temp);
        }

      // find time before temp
      here = m_listHead;
      for (gopher = m_listHead; gopher && gopher != temp; gopher = gopher->next)
        {
          if (gopher->labtime != 0.0) here = gopher;
        }

      Out2File(here);
    }
  while (entryCount >= MAXSIZE);

  ClearList();
  CloseFiles();

  // rename last one
  if (!m_analysis && (m_currentID == WATSON || m_currentID == LASER || m_currentID == HORITA))
    {
      sprintf(m_newFile,"%s%d.txt",m_outputFileName, m_fileNumber-1);
      oldFile = strdup(m_newFile);
      if (oldFile)
        {
          sprintf(m_newFile,"%s%d_%1.1lf-%1.1lf.txt",m_outputFileName, m_fileNumber-1, m_startLab, m_endLab);
          rename(oldFile, m_newFile);
          delete[] oldFile;
        }
    }

  return true;
}


bool CFlightData::ReadRawHorita()
{
  m_currentID = HORITA_RR;
  return CleanRawFile();
}

void CFlightData::ProcessUserData(int frame, int u1, int u2, int u3, int u4)
{
  if (!fpOUT) return;
  switch (frame)
    {
    case 0:
      fprintf(fpOUT,"date = %d/%d/%d\n",u3, u4, u2);
      break;
    case 2:
      fprintf(fpOUT,"Satellite #8 PRN = %d; Number of satellites tracked = %d\n",u2,u3);
      break;
    case 3:
      fprintf(fpOUT,"Satellite #1 PRN = %d; Satellite #2 PRN = %d\n", u4, u3);
      break;
    case 4:
      fprintf(fpOUT,"Position Fix Mode, TSIP #%d:",u2);
      if (u2 == 04) fprintf(fpOUT,"AUTO 3D; ");
      if (u2 == 03) fprintf(fpOUT,"AUTO 2D; ");
      if (u2 == 14) fprintf(fpOUT,"MANUAL 3D; ");
      if (u2 == 13) fprintf(fpOUT,"MANUAL 2D; ");
      fprintf(fpOUT,"Satellite #3 PRN = %d, Satellite #4 PRN = %d\n",u4,u3);
      break;
    case 5:
      fprintf(fpOUT,"Time of Position Fix = %d:%d:%d\n",u2, u3, u4);
      break;
    case 12:
      fprintf(fpOUT,"Satellite #7 PRN = %d; Satellite #6 PRN = %d; ",u2,u3);
      fprintf(fpOUT,"Satellite #5 PRN = %d\n",u4);
      break;
    case 21:
      fprintf(fpOUT,"%s Latitude Degrees = %d, Minutes = %d\n",(u2?"South":"North"),u3,u4);
      break;
    case 22:
      fprintf(fpOUT,"Latitude fraction of minutes = %d%s%d\n",u3,(u4>9?"":"0"),u4);
      break;
    case 23:
      fprintf(fpOUT,"%s Longitude Degrees = %d%s%d, Minutes = %d\n",(u2>9?"West":"East"), u2%10,(u3>9?"":"0"),u3,u4);
      break;
    case 24:
      fprintf(fpOUT,"Longitude fraction of minutes = %d%s%d; ",u3,(u4>9?"":"0"),u4);
      fprintf(fpOUT,"Altitude is %s Sea Level\n",(u2%10?"Below":"Above"));
      break;
    case 25:
      fprintf(fpOUT,"Altitude = %d%s%d%s%d\n",u2%10,(u3>9?"":"0"),u3,(u4>9?"":"0"),u4);
      break;
    case 26:
      fprintf(fpOUT,"Upward Velocity (feet/second) = %s%d%s%d%d.%d\n",(u2/10?"-":"+"),
              u2%10,(u3>9?"":"0"),u3,u4/10,u4%10);
      break;
    case 27:
      fprintf(fpOUT,"East Velocity (mph) = %s%d%s%d%d.%d\n",(u2/10?"-":"+"), u2%10,(u3>9?"":"0"),u3,u4/10,u4%10);
      break;
    case 28:
      fprintf(fpOUT,"North Velocity (mph) = %s%d%s%d%d.%d\n",(u2/10?"-":"+"), u2%10,(u3>9?"":"0"),u3,u4/10,u4%10);
      break;
    default:
      fprintf(fpOUT,"\n");
    }
  return;
}

bool CFlightData::Composer()
{
  ReadStatus statHor, statLas, statWat;
  int i, size, fileLineCount = 0;
  double aTime, syncTime = 0.0;
  LAB *temp, *here, *marker;

  // initialize
  ClearList();
  m_masterDate[0] = '\0';
  m_fileNumber = 0;

  /////// OPEN INPUT AND OUTPUT FILES ///////
  if (m_composerFlags[cLASER_FILE])
    {
      fpLAS = fopen(m_inputLaserFile,"r");
      if (!fpLAS)
        {
          Out2User("Cannot Open Raw Laser File");
          CloseFiles();
          return false;
        }
      syncTime = CheckSort(fpLAS);
      if (syncTime == 0.0)
        {
          Out2User("The timestamps in the Laser File are not sequential");
          CloseFiles();
          return false;
        }
    }

  if (m_composerFlags[cWATSON_FILE])
    {
      fpWAT = fopen(m_inputWatsonFile,"r");
      if (!fpWAT)
        {
          Out2User("Cannot Open Raw Watson File");
          CloseFiles();
          return false;
        }
      aTime = CheckSort(fpWAT);
      if (aTime == 0.0)
        {
          Out2User("The timestamps in the Watson File are not sequential");
          CloseFiles();
          return false;
        }
      if (aTime > syncTime) syncTime = aTime;
    }

  fpHOR = fopen(m_inputHoritaFile,"r");
  if (!fpHOR)
    {
      Out2User("Cannot Open Raw Horita File");
      CloseFiles();
      return false;
    }
  aTime = CheckSort(fpHOR);
  if (aTime == 0.0)
    {
      Out2User("The timestamps in the Horita File are not sequential");
      CloseFiles();
      return false;
    }
  if (aTime > syncTime) syncTime = aTime;

  // prepare output file
  for (i = strlen(m_outputFileName); i >= 0; i--)
    {
      if (m_outputFileName[i] == '.')
        {
          m_outputFileName[i] = '\0';
          break;
        }
      if (m_outputFileName[i] == '\\' || m_outputFileName[i] == '/') break;
    }

  // load the first Horita value beyond syncTime with frame zero
  temp = GetLab();
  if (!temp)
    {
      Out2User("Memory Error Queuing the Horita");
      CloseFiles();
      return false;
    }

  do
    {
      if (HoritaRead(temp) != rGOOD)
        {
          delete temp;
          Out2User("Error scanning Horita File for first Matching Time");
          CloseFiles();
          return false;
        }
    }
  while(temp->labtime < syncTime || temp->frame != 0);

  // add to list and adjust syncTime
  temp->next = m_masterList;
  m_masterList = temp;
  syncTime = temp->labtime;

  // queue the Laser to match the Horita time (if a Laser File was entered)
  if (m_composerFlags[cLASER_FILE])
    {
      if (!QueueLasWat(true, syncTime)) return false;
    }

  // queue the Watson to match the Horita Time (if a Watson File was entered)
  if (m_composerFlags[cWATSON_FILE])
    {
      if (!QueueLasWat(false, syncTime)) return false;
    }

  // AT THIS POINT ALL THREE(?) FILES ARE AT A MATCHING LABTIME.

  // load list
  // load horita
  size = MASTERSIZE * 30;
  if (size < 90) size = 90;
  aTime = syncTime;

  do
    {
      for (i = 0; i < size; i++)
        {
          temp = GetLab();
          if (!temp)
            {
              Out2User("Memory Error Loading Horita Data");
              CloseFiles();
              ClearList();
              return false;
            }
          statHor = HoritaRead(temp);
          if (statHor == rFILE_END)
            {
              delete temp;
              break;
            }
          if (statHor == rBAD_SCAN)
            {
              delete temp;
              Out2User("Error scanning Horita File (2)");
              CloseFiles();
              return false;
            }
          // add to list
          syncTime = temp->labtime;
          temp->next = m_masterList;
          m_masterList = temp;
        }

      // load laser / fill in laser value
      if (m_composerFlags[cLASER_FILE])
        {
          if (!LoadLasWat(true, syncTime, &statLas)) return false;
        }

      // load watson
      if (m_composerFlags[cWATSON_FILE])
        {
          if (!LoadLasWat(false, syncTime, &statWat)) return false;
        }

      // fill in any blanks, and prepare for sort
      for (i = 0, temp = m_masterList; temp; i++, temp = temp->next)
        {
          // if no laser files, fill in laser data
          if (!m_composerFlags[cLASER_FILE])
            {
              temp->laser = m_laserValue;
              temp->fLaser = true;
            }
        }

      // process list - sort
      m_masterList = MergeSort(m_masterList, i);

      // combine same-labtime data
      for (here = m_masterList; here && here->next; here = here->next)
        {
          if (here->labtime == here->next->labtime)
            {
              temp = here->next;
              if (!here->laser) here->laser = temp->laser;
              if (!here->scatter) here->scatter = temp->scatter;
              if (!here->watsValid) here->watsValid = temp->watsValid;
              if (!here->roll) here->roll = temp->roll;
              if (!here->pitch) here->pitch = temp->pitch;
              if (!here->heading) here->heading = temp->heading;
              if (!here->hour) here->hour = temp->hour;
              if (!here->min) here->min = temp->min;
              if (!here->sec) here->sec = temp->sec;
              if (!here->frame) here->frame = temp->frame;
              if (!here->u1) here->u1 = temp->u1;
              if (!here->u2) here->u2 = temp->u2;
              if (!here->u3) here->u3 = temp->u3;
              if (!here->u4) here->u4 = temp->u4;
              if (!here->theDate[0]) strcpy(here->theDate,temp->theDate);
              if (!here->fLaser) here->fLaser = temp->fLaser;
              if (!here->fWatson) here->fWatson = temp->fWatson;
              if (!here->fHorita) here->fHorita = temp->fHorita;
              here->next = temp->next;
              delete temp;
            }
        }

      // compute UTM from Lat/Long Horita code (function returns last good Horita)
      here = CorrectGPS();
      marker = here;

      // find the next watson and break the string
      while (marker && m_composerFlags[cWATSON_FILE] && marker->watsValid == '\0') marker = marker->next;

      if (marker)
        {
          temp = marker->next;
          marker->next = 0;
        }

      // incorporate locational offset of GPS antenna to cameras
      OffsetGPS();

      // interpolate all GPS and Watson data
      InterGPS();
      if (m_composerFlags[cWATSON_FILE]) InterWat();

      // correct the heading using GPS locations
      if (!m_composerFlags[cORIG_HEADING]) CorrectHeading();

      // compute ground coordinates
      GPFinder();

      // put the entire string back together
      if (marker) marker->next = temp;

      // NOTE: 'here' contains the last good Horita, who's values may or may not have been calculated
      // since the correction procedure need the full second to gather all the GPS data for THIS zeroth frame.
      // So, back up one full second...
      for (marker = m_masterList; marker != here; marker = marker->next)
        {
          if (marker->fHorita == 2) temp = marker;
        }
      here = temp;

      // now output list
      for (temp = m_masterList; temp && temp != here; temp = temp->next)
        {
          if (temp->labtime < aTime) continue;								// removes entries prior to beginning point
          if (m_composerFlags[cLASER_ONLY] && !temp->fLaser) continue;		// only output data corresponding to laser data
          if (m_composerFlags[cNO_ZEROS] && temp->gndnorth == 0.0) continue;	// only output nonzero laser data

          if (!m_composerFlags[cOUTPUT_SINGLE] && fpOUT && fileLineCount > 64000)
            {
              fclose(fpOUT);
              fileLineCount = 0;
              fpOUT = 0;
            }
          if (!fpOUT)
            {
              // create output filename
              if (!m_composerFlags[cOUTPUT_SINGLE])
                {
                  for (marker = temp; marker && (!marker->hour && !marker->min); marker = marker->next);
                  if (!marker) marker = temp;
                  sprintf(m_newFile,"%s%d%s%d%s%d%s%d.txt",m_outputFileName, marker->hour, (marker->min<10)?"-0":"-", marker->min,
                          (marker->sec<10)?"-0":"-", marker->sec, (marker->frame<10)?"-0":"-", marker->frame);
                }
              else sprintf(m_newFile,"%s.txt",m_outputFileName);

              fpOUT = fopen (m_newFile,"w");
              if (!fpOUT)
                {
                  ClearList();
                  CloseFiles();
                  Out2User("Could not open output File");
                }
              fprintf(fpOUT,"labtime      laser/scat   tilt tip heading   timecode    alt  ");
              fprintf(fpOUT,"a/c_North  a/c_East gnd_alt gnd_North gnd_East Zone Date %s\n\n",
                      m_markLines?"Inst":"");
              Out2User("Writing to output file...");
            }

          fprintf(fpOUT,"%9.1lf ",temp->labtime);
          fprintf(fpOUT," %7.1lf %3d ",temp->laser, temp->scatter);
          fprintf(fpOUT,"%c %5.1f %4.1f %4.1f ", temp->watsValid, temp->roll, temp->pitch, temp->heading);
          fprintf(fpOUT," %d%s%d%s",temp->hour, temp->min<10?":0":":", temp->min, temp->sec<10?":0":":");
          fprintf(fpOUT,"%d%s%d ", temp->sec, temp->frame<10?":0":":", temp->frame);
          fprintf(fpOUT,"%6.1lf %9.1lf %9.1lf ",temp->alt, temp->north, temp->east);
          fprintf(fpOUT,"%6.1lf %9.1lf %9.1lf",temp->gndalt, temp->gndnorth, temp->gndeast);
          fprintf(fpOUT," %s %s",temp->zone, temp->theDate);
          if (m_markLines)
            {
              if (temp->fLaser) fprintf(fpOUT," L");
              if (temp->fWatson) fprintf(fpOUT," W");
              if (temp->fHorita) fprintf(fpOUT," H");
            }
          fprintf(fpOUT,"\n");
          if (!m_composerFlags[cOUTPUT_SINGLE]) fileLineCount++;
        }

      // clear the printed data nodes
      while (m_masterList != here)
        {
          temp = m_masterList;
          m_masterList = m_masterList->next;
          delete temp;
        }
      // repeat
    }
  while (statHor != rFILE_END && statLas != rFILE_END && statWat != rFILE_END);

  ClearList();
  CloseFiles();
  Out2User("Done");
  return true;  
}

bool CFlightData::GenPlot()
{
  int i;
  float airNorth, airEast, gndAlt, gndNorth, gndEast, oldNorth = 0.0, oldEast = 0.0;
  fpHOR = fopen(m_inputHoritaFile,"r");
  if (!fpHOR)
    {
      Out2User("Cannot Open Meta-File Entered");
      return false;
    }

  fpOUT = fopen(m_outputFileName,"w");
  if (!fpOUT)
    {
      Out2User("Cannot Open Output File");
      CloseFiles();
      return false;
    }

  while (fgets(m_newFile, 255, fpHOR))
    {
      // labtime laser scat watsvalid roll pitch heading TC alt N E gndalt gndN gndE
      i = sscanf(m_newFile,"%*s %*s %*s %*s %*s %*s %*s %*s %*s %f %f %f %f %f",
                     &airNorth, &airEast, &gndAlt, &gndNorth, &gndEast);
      if (i < 5)
        {
          printf("Scanning Error: %s\n",m_newFile);
          continue;
        }

      switch(pType)
        {
        case PLOT_ALT:
          if (gndAlt < 0.1) break;

          if (oldNorth == 0.0)
            {
              oldNorth = gndNorth;
              oldEast = gndEast;
            }
          else
            {
              double y = (double)(gndNorth - oldNorth);
              double x = (double)(gndEast - oldEast);
              double distance = sqrt(y*y + x*x);
              fprintf(fpOUT,"%1.1lf\t%1.1lf\n", distance, gndAlt);
            }
          break;
        case PLOT_GROUND:
          if (gndEast < 0.1) break;
          fprintf(fpOUT,"%1.1f\t1.1f\n", gndEast, gndNorth);
          break;
        case PLOT_PLANE:
          if (airEast < 0.1) break;
          fprintf(fpOUT,"%1.1f\t1.1f\n", airEast, airNorth);
          break;
        }
    }
  CloseFiles();
  return true;
}

//////////////////////////////////////////////////////////////////////
// This procedure checks a file to ensure the labtime is sequential //
//////////////////////////////////////////////////////////////////////
double CFlightData::CheckSort(FILE * fp)
{
  char buf[64];
  double time2, time1 = 0.0;
  int status;

  rewind(fp);
  while (fgets(buf,63,fp))
    {
      status = sscanf(buf,"%lf",&time2);
      if (status != 1)
        {
          printf("%s\n",buf);
          continue;
        }
      if (time1 > time2) return (0.0);
      time1 = time2;
    }
  rewind(fp);
  if (fgets(buf,63,fp)) sscanf(buf,"%lf",&time2);
  else time2 = 0.0;
  rewind(fp);
  return time2;
}

ReadStatus CFlightData::HoritaRead(LAB * temp)
{
  if (fgets(m_newFile,99,fpHOR))
    {
      int i = sscanf(m_newFile,"%lf %*s %hd %hd %hd %hd %*s %hd %hd %hd %hd",&(temp->labtime), &(temp->frame),
                     &(temp->sec), &(temp->min), &(temp->hour), &(temp->u4), &(temp->u3), &(temp->u2), &(temp->u1));
      if (i < 5) return rBAD_SCAN;
      temp->fHorita = 1;
      return rGOOD;
    }
  return rFILE_END;
}

ReadStatus CFlightData::LaserRead(LAB * temp)
{
  if (fgets(m_newFile,99,fpLAS))
    {
      int i = sscanf(m_newFile,"%lf %lf %d",&(temp->labtime), &(temp->laser), &(temp->scatter));
      if (i < 3) return rBAD_SCAN;
      temp->fLaser = true;
      return rGOOD;
    }
  return rFILE_END;
}

ReadStatus CFlightData::WatsonRead(LAB * temp)
{
  if (fgets(m_newFile,99,fpWAT))
    {
      int i = sscanf(m_newFile,"%lf %c %f %f %f", &(temp->labtime), &(temp->watsValid), &(temp->roll), &(temp->pitch), &(temp->heading));
      if (i < 5) return rBAD_SCAN;
      temp->roll += m_rollOffset;
      temp->pitch += m_pitchOffset;
      temp->fWatson = true;
      return rGOOD;
    }
  return rFILE_END;
}

LAB * CFlightData::CorrectGPS()
{
  LAB *temp, *here = 0;
  bool count = false;
  char latbuf[12], longbuf[12], altbuf[12];
  int hour = -1, min = 1, sec = 1, latdeg = 0, latmin = 0, longdeg = 0, longmin = 0, latsign = 1, longsign = 1;
  double Lat = 0.0, Long = 0.0, UTMnorth = 0.0, UTMeast = 0.0;

  for (temp = m_masterList; temp; temp = temp->next)
    {
      // find first frame of this timecode sequence
      if (!here && temp->fHorita && temp->frame < 5 && (temp->hour != hour || temp->min != min || temp->sec != sec))
        {
          here = temp;
          hour = temp->hour;
          min = temp->min;
          sec = temp->sec;
          if (temp->frame == 0)
            sprintf(m_masterDate,"%d-%d-%d",temp->u4,temp->u3,temp->u2);
        }

      if (temp->frame == 21 && here && temp->hour == hour && temp->min == min && temp->sec == sec)
        {
          latdeg = temp->u3;
          latmin = temp->u4;
          if (temp->u2) latsign = -1;
          else latsign = 1;
        }

      if (temp->frame == 22 && here && temp->hour == hour && temp->min == min && temp->sec == sec)
        {
          sprintf(latbuf,"%d.%s%d%s%d", latmin, (temp->u3<10?"0":""), temp->u3, (temp->u4<10?"0":""), temp->u4);
          sscanf(latbuf,"%lf",&Lat);
          Lat = (Lat/60.0 + ((double)latdeg)) * ((double)latsign);
        }

      if (temp->frame == 23 && here && temp->hour == hour && temp->min == min && temp->sec == sec)
        {
          longdeg = (temp->u2 % 10) * 100 + temp->u3;
          longmin = temp->u4;
          if (temp->u2 > 9 ) longsign = -1;
          else longsign = 1;
        }

      if (temp->frame == 24 && here && temp->hour == hour && temp->min == min && temp->sec == sec)
        {
          sprintf(longbuf,"%d.%s%d%s%d", longmin, (temp->u3<10?"0":""), temp->u3, (temp->u4<10?"0":""), temp->u4);
          sscanf(longbuf,"%lf",&Long);
          Long = (Long/60.0 + ((double)longdeg)) * ((double)longsign);
          LLtoUTM( Lat, Long, &UTMnorth, &UTMeast, here->zone);
          if (here->north == 0.0 && here->east == 0.0)
            {
              here->north = UTMnorth;
              here->east = UTMeast;
              here->fHorita = 2;
            }
        }

      if (temp->frame == 25 && here && temp->hour == hour && temp->min == min && temp->sec == sec)
        {
          sprintf(altbuf,"%d%s%d%s%d", temp->u2, (temp->u3<10?"0":""), temp->u3, (temp->u4<10?"0":""), temp->u4);
          sscanf(altbuf,"%lf",&(here->alt));

          // convert altitude from feet to meters
          here->alt = here->alt * .3048;
          count = true;
        }

      if (count || (here && temp->labtime > here->labtime + 1200.0))
        {
          temp = here;
          here = 0;
          latdeg = 0;
          latmin = 0;
          longdeg = 0;
          longmin = 0;
          count = false;
        }
    }
  return here;
}

LAB * CFlightData::MergeSort(LAB * lsta, long count)
{
  LAB *lstb, *temp;
  long i;
  if (count > 1)
    {
      temp = lsta;
      for (i = 0; i < (count/2)-1; i++)
        temp = temp->next;
      lstb = temp->next;
      temp->next = 0;
      lsta = MergeSort(lsta,count/2);
      lstb = MergeSort(lstb,(count - count/2));
      lsta = Merge(lsta,lstb);
    }
  return (lsta);
}

LAB * CFlightData::Merge(LAB * lsta, LAB * lstb)
{
  LAB *lst, *temp;

  if (lsta == 0) return (lstb);
  if (lstb == 0) return (lsta);

  if (lsta->labtime < lstb->labtime)
    {
      lst = lsta;
      lsta = lsta->next;
    }
  else
    {
      lst = lstb;
      lstb = lstb->next;
    }
  temp = lst;

  while (lsta && lstb)
    {
      if (lsta->labtime < lstb->labtime)
        {
          temp->next = lsta;
          lsta = lsta->next;
        }
      else
        {
          temp->next = lstb;
          lstb = lstb->next;
        }
      temp = temp->next;
    }

  if (lsta) temp->next = lsta;
  else temp->next = lstb;
  return (lst);
}

/*******************************************/
/* Table for computing various projections */ 
/*******************************************/
/*	"Airy"					EquatorialRadius = 6377563.0; eccentricitySquared = 0.00667054;
	"Australian National"	EquatorialRadius = 6378160.0; eccentricitySquared = 0.006694542;
	"Bessel 1841"			EquatorialRadius = 6377397.0; eccentricitySquared = 0.006674372;
	"Bessel 1841 (Nambia)"	EquatorialRadius = 6377484.0; eccentricitySquared = 0.006674372;
	"Clarke 1866"			EquatorialRadius = 6378206.0; eccentricitySquared = 0.006768658;
	"Clarke 1880"			EquatorialRadius = 6378249.0; eccentricitySquared = 0.006803511;
	"Everest"				EquatorialRadius = 6377276.0; eccentricitySquared = 0.006637847;
	"Fischer 1960 (Mercury)"EquatorialRadius = 6378166.0; eccentricitySquared = 0.006693422;
	"Fischer 1968"			EquatorialRadius = 6378150.0; eccentricitySquared = 0.006693422;
	"GRS 1967"				EquatorialRadius = 6378160.0; eccentricitySquared = 0.006694605;
	"GRS 1980"				EquatorialRadius = 6378137.0; eccentricitySquared = 0.00669438;
	"Helmert 1906"			EquatorialRadius = 6378200.0; eccentricitySquared = 0.006693422;
	"Hough"					EquatorialRadius = 6378270.0; eccentricitySquared = 0.00672267;
	"International"			EquatorialRadius = 6378388.0; eccentricitySquared = 0.00672267;
	"Krassovsky"			EquatorialRadius = 6378245.0; eccentricitySquared = 0.006693422;
	"Modified Airy"			EquatorialRadius = 6377340.0; eccentricitySquared = 0.00667054;
	"Modified Everest"		EquatorialRadius = 6377304.0; eccentricitySquared = 0.006637847;
	"Modified Fischer 1960"	EquatorialRadius = 6378155.0; eccentricitySquared = 0.006693422;
	"South American 1969"	EquatorialRadius = 6378160.0; eccentricitySquared = 0.006694542;
	"WGS 60"				EquatorialRadius = 6378165.0; eccentricitySquared = 0.006693422;
	"WGS 66"				EquatorialRadius = 6378145.0; eccentricitySquared = 0.006694542;
	"WGS-72"				EquatorialRadius = 6378135.0; eccentricitySquared = 0.006694318;
	"WGS-84"				EquatorialRadius = 6378137.0; eccentricitySquared = 0.00669438;
*/
#include <math.h>
#include <stdlib.h>
void CFlightData::LLtoUTM(double Lat, double Long, double * UTMNorthing, double * UTMEasting, char * UTMZone)
{
  double a, eccSquared, LongOrigin, eccPrimeSquared, N, T, C, A, M, k0 = 0.9996;
  double LatRad, LongRad, LongOriginRad;

  // values taken from table above
  a = 6378137.0;
  eccSquared = 0.00669438;
  LatRad = Lat*r_pie/180.0;
  LongRad = Long*r_pie/180.0;

  if(Long > -6 && Long <= 0) LongOrigin = -3;//arbitrarily set origin at 0 longitude to 3W
  else if(Long < 6 && Long > 0) LongOrigin = 3;
  else LongOrigin = (double) (((int)(Long/6))*6+3*(((int)(Long/6))/abs(((int)(Long/6)))));

  LongOriginRad = LongOrigin * r_pie / 180.0;

  //compute the UTM Zone from the latitude and longitude
  sprintf(UTMZone, "%d%c", ((int)(((Long + 180)/6) + 1)), UTMLetterDesignator(Lat));

  eccPrimeSquared = (eccSquared)/(1-eccSquared);

  N = a/sqrt(1-eccSquared*sin(LatRad)*sin(LatRad));
  T = tan(LatRad)*tan(LatRad);
  C = eccPrimeSquared*cos(LatRad)*cos(LatRad);
  A = cos(LatRad)*(LongRad-LongOriginRad);

  M = a*((1 - eccSquared/4 - 3*eccSquared*eccSquared/64 - 5*eccSquared*eccSquared*eccSquared/256)
	 *LatRad - (3*eccSquared/8 + 3*eccSquared*eccSquared/32
		    + 45*eccSquared*eccSquared*eccSquared/1024)*sin(2*LatRad)
	 + (15*eccSquared*eccSquared/256 + 45*eccSquared*eccSquared*eccSquared/1024)*sin(4*LatRad) 
	 - (35*eccSquared*eccSquared*eccSquared/3072)*sin(6*LatRad));

  *UTMEasting = (double)(k0*N*(A+(1-T+C)*A*A*A/6 + (5-18*T+T*T+72*C-58*eccPrimeSquared)*A*A*A*A*A/120) + 500000.0);

  *UTMNorthing = (double)(k0*(M+N*tan(LatRad)*(A*A/2+(5-T+9*C+4*C*C)*A*A*A*A/24 + (61-58*T+T*T+600*C-330*eccPrimeSquared)*A*A*A*A*A*A/720)));

  if(Lat < 0) *UTMNorthing += 10000000.0; //10000000 meter offset for southern hemisphere

}

//////////////////////////////////////////////////////////////////////////////////////
// This routine determines the correct UTM letter designator for the given latitude //
// returns 'Z' if latitude is outside the UTM limits of 80N to 80S                  //
// Written by Chuck Gantz- chuck.gantz@globalstar.com                               //
//////////////////////////////////////////////////////////////////////////////////////
char CFlightData::UTMLetterDesignator(double Lat)
{
  char LetterDesignator;

  if((80 >= Lat) && (Lat > 72)) LetterDesignator = 'X';
  else if((72 >= Lat) && (Lat > 64)) LetterDesignator = 'W';
  else if((64 >= Lat) && (Lat > 56)) LetterDesignator = 'V';
  else if((56 >= Lat) && (Lat > 48)) LetterDesignator = 'U';
  else if((48 >= Lat) && (Lat > 40)) LetterDesignator = 'T';
  else if((40 >= Lat) && (Lat > 32)) LetterDesignator = 'S';
  else if((32 >= Lat) && (Lat > 24)) LetterDesignator = 'R';
  else if((24 >= Lat) && (Lat > 16)) LetterDesignator = 'Q';
  else if((16 >= Lat) && (Lat > 8)) LetterDesignator = 'P';
  else if(( 8 >= Lat) && (Lat > 0)) LetterDesignator = 'N';
  else if(( 0 >= Lat) && (Lat > -8)) LetterDesignator = 'M';
  else if((-8>= Lat) && (Lat > -16)) LetterDesignator = 'L';
  else if((-16 >= Lat) && (Lat > -24)) LetterDesignator = 'K';
  else if((-24 >= Lat) && (Lat > -32)) LetterDesignator = 'J';
  else if((-32 >= Lat) && (Lat > -40)) LetterDesignator = 'H';
  else if((-40 >= Lat) && (Lat > -48)) LetterDesignator = 'G';
  else if((-48 >= Lat) && (Lat > -56)) LetterDesignator = 'F';
  else if((-56 >= Lat) && (Lat > -64)) LetterDesignator = 'E';
  else if((-64 >= Lat) && (Lat > -72)) LetterDesignator = 'D';
  else if((-72 >= Lat) && (Lat > -80)) LetterDesignator = 'C';
  else LetterDesignator = 'Z'; //Error flag to show that the Latitude is outside the UTM limits

  return LetterDesignator;

}

void CFlightData::OffsetGPS()
{
  double nd, ed, hyp, noff, eoff, aoff;
  LAB *temp, *here;

  if (!m_cameraForward && !m_cameraSide && !m_cameraVertical) return;

  // input is in feet - convert to meters
  noff = .3048 * (double)m_cameraForward;
  eoff = .3048 * (double)m_cameraSide;
  aoff = .3048 * (double)m_cameraVertical;

  for (here = m_masterList; here;)
    {
      if (here->fHorita == 2)
        {
          temp = here->next;
          while (temp && temp->fHorita != 2) temp = temp->next;
          if (temp == 0) return;
          if (temp->oldnorth)
            {
              here = here->next;
              continue;
            }
          if (here->oldnorth) nd = temp->north - here->oldnorth;
          else nd = temp->north - here->north;
          if (here->oldeast) ed = temp->east - here->oldeast;
          else ed = temp->east - here->east;
          if (nd || ed) hyp = sqrt (nd * nd + ed * ed);
          else hyp = 0;
          if (here->oldnorth == 0 && hyp)
            {
              here->oldnorth = here->north;
              here->north = here->north + ((noff * nd) / hyp);
              here->oldeast = here->east;
              here->east = here->east + ((eoff * ed) / hyp);
              here->oldalt = here->alt;
              here->alt += aoff;
            }
          if (temp->oldnorth == 0 && hyp)
            {
              temp->oldnorth = temp->north;
              temp->north = temp->north + ((noff * nd) / hyp);
              temp->oldeast = temp->east;
              temp->east = temp->east + ((eoff * ed) / hyp);
              temp->oldalt = temp->alt;
              temp->alt += aoff;
            }
          here = temp;
        }
      else here = here->next;
    }
  return;
}


void CFlightData::InterGPS()
{
  double nd, ed, ad, n1, e1, a1, t1, t2, tr;
  int hour, min, sec, frame;
  char c[5], thedate[10];
  LAB *temp, *here;

  thedate[0] = '\0';
  for (here = m_masterList; here; here = here->next)
    {
      if (here->fHorita == 2)
        {
          t1 = here->labtime;
          n1 = here->north;
          e1 = here->east;
          a1 = here->alt;
          hour = here->hour;
          min = here->min;
          sec = here->sec;
          frame = here->frame;
          strcpy (c, here->zone);
          if (here->theDate[0]) strcpy (thedate, here->theDate);
          else
            {
              strcpy (thedate, m_masterDate);
              if (m_masterDate[0]) strcpy (here->theDate, m_masterDate);
            }

          temp = here->next;
          while (temp && temp->fHorita != 2)
            temp = temp->next;
          if (temp == 0) return;

          if (strcmp(temp->zone, here->zone) != 0) continue;
          nd = temp->north - n1;
          ed = temp->east - e1;
          ad = temp->alt - a1;
          t2 = temp->labtime - t1;
          while (here->next != temp)
            {
              here = here->next;
              if (here->hour == 0)
                {
                  here->hour = hour;
                  here->min = min;
                  here->sec = sec;
                  here->frame = frame;
                }
              else
                {
                  frame = here->frame;
                }
              strcpy (here->zone, c);
              if (thedate[0]) strcpy (here->theDate, thedate);
              tr = (here->labtime - t1) / t2;
              here->north = (tr * nd) + n1;
              here->east = (tr * ed) + e1;
              here->alt = (tr * ad) + a1;
            }
        }
    }
  return;
}

void CFlightData::InterWat()
{
  float rolld, pitchd, headd, roll1, pitch1, head1;
  double t1, t2, tr;
  char c;
  LAB *temp, *here;

  for (here = m_masterList; here; here = here->next)
    {
      if (here->fWatson)
        {
          t1 = here->labtime;
          roll1 = here->roll;
          pitch1 = here->pitch;
          head1 = here->heading;
          c = here->watsValid;

          temp = here->next;
          while (temp && !temp->fWatson)
            temp = temp->next;
          if (temp == 0) return;

          if (c < temp->watsValid) c = temp->watsValid;
          rolld = temp->roll - roll1;
          pitchd = temp->pitch - pitch1;
          headd = temp->heading - head1;
          t2 = temp->labtime - t1;
          while (here->next != temp)
            {
              here = here->next;
              here->watsValid = c;
              tr = (here->labtime - t1) / t2;
              here->roll = ((float) (tr * ((double) rolld))) + roll1;
              here->pitch = ((float) (tr * ((double) pitchd))) + pitch1;
              here->heading = ((float) (tr * ((double) headd))) + head1;
            }
        }
    }
  return;
}

void CFlightData::CorrectHeading()
{
  double nd, ed, n1, e1, n2, e2, head;
  LAB *temp, *here;

  for (here = m_masterList; here; here = here->next)
    {
      if (here->fHorita == 2)
        {
          n1 = here->north;
          e1 = here->east;

          temp = here->next;
          while (temp && temp->fHorita != 2)
            temp = temp->next;
          if (temp == 0) return;
          if (strcmp(temp->zone, here->zone) != 0) continue;

          n2 = temp->north;
          e2 = temp->east;
          nd = n2 - n1;
          ed = e2 - e1;

          // COMPUTE HEADING
          if (ed == 0)
            {
              if (n1 > n2) head = 180.0;
              else head = 0.0;
            }
          else if (nd == 0)
            {
              if (e1 > e2) head = 270.0;
              else head = 90.0;
            }
          else
            {
              head = atan(ed/nd); // return is in radians
              head *= 180/3.14159265358;    // convert to degrees
              if (ed < 0 && nd > 0) head = 360.0 + head;
              if (nd < 0) head = 180.0 + head;
            }

          here->oldheading = here->heading;
          here->heading = (float) head; 
          while (here->next != temp)
            {
              here = here->next;
              here->oldheading = here->heading;
              here->heading = (float) head;
            }
        }
    }
  return;
}

void CFlightData::GPFinder()
{
  bool flag;
  double north1, east1, north2, east2, t1, td, tr, laser;
  double pitch, tilt, n_plane, e_plane;
  double a, b, c, d, altitude, flightdis, wingdis;
  double rollquad, pitchquad, x1, x2, y1, y2, xf, yf;
  LAB *point1, *point2;

  for (point1 = m_masterList; point1; point1 = point1->next)
    {
      if (point1->fHorita == 2) // authentic GPS data reading
        {
          north1 = point1->north;
          east1 = point1->east;
          t1 = point1->labtime;

          // find next authentic GPS reading
          for(point2 = point1->next; point2 && point2->fHorita != 2; point2 = point2->next);
          if (!point2) return;

          north2 = point2->north;
          east2 = point2->east;
          td = point2->labtime - t1;
          flag = true;
          while (flag)
            {
              if (point1->laser && point1->alt > 5)	// ensure all values available for ground computation
                {
                  tr = (point1->labtime - t1) / td;	// calculate time ratio
                  a = (double)(point1->pitch);		// process angle data
                  b = (double)(point1->roll);
                  if(a >= 90 || a <= -90 || b >= 90 || b <= -90) goto skipit;
                  // determine quadrant
                  if (a < 0) pitchquad = -1.0;
                  else pitchquad = 1.0;
                  if (b < 0) rollquad = -1.0;
                  else rollquad = 1.0;
                  // make angles positive and convert degrees to radians
                  pitch = (a * pitchquad * 2.0 * r_pie) / 360.0;
                  tilt  = (b * rollquad * 2.0 * r_pie) / 360.0;

                  a = tan(pitch);
                  b = tan(tilt);

                  if (m_composerFlags[cLASER_ALT])		// here laser value is average plane local altitude
                    {
                      altitude = point1->laser * .3048;
                    }
                  else if (m_composerFlags[cLASER_GND])	// here laser value is average ground altitude
                    {
                      altitude = point1->alt - point1->laser;
                    }
                  else // cLASER_AVE, cLASER_FILE			// here laser value is laser value (average/measured)
                    {
                      // convert feet to meters
                      laser = point1->laser * .3048;

                      // compute vertical distance from plane to ground (local altitude)
                      c = (a * a) + (b * b) + 1;
                      d = laser * laser;
                      if (d == 0.0) altitude = 0;
                      else altitude = sqrt(d / c);
                    }

                  flightdis = a * altitude * pitchquad;
                  wingdis = b * altitude * rollquad;
                  n_plane = point1->north;
                  e_plane = point1->east;
                  /* Northeast Direction */
                  if ((north2 >= north1) && (east2 >= east1))
                    {
                      a = east2 - e_plane;
                      b = north2 - n_plane;
                      c = (a * a) + (b * b);
                      if (c == 0)
                        {
                          xf = east2;
                          yf = north2;
                        }
                      else
                        {
                          d = sqrt(c);
                          x1 = (a / d) * flightdis;
                          y1 = (b / d) * flightdis;
                          x2 = ( b / d) * wingdis;
                          y2 = ( a / d) * wingdis;
                          xf = e_plane + (x1 - x2);
                          yf = n_plane + (y1 + y2);
                        }
                    }
                  /* Northwest Direction */
                  if ((north2 >= north1) && (east2 < east1))
                    {
                      a = e_plane - east2;
                      b = north2 - n_plane;
                      c = (a * a) + (b * b);
                      if (c == 0)
                        {
                          xf = east2;
                          yf = north2;
                        }
                      else
                        {
                          d = sqrt(c);
                          x1 = (a / d) * flightdis;
                          y1 = (b / d) * flightdis;
                          x2 = ( b / d) * wingdis;
                          y2 = ( a / d) * wingdis;
                          xf = e_plane - (x1 + x2);
                          yf = n_plane + (y1 - y2);
                        }
                    }
                  /* Southeast Direction */
                  if ((north2 < north1) && ( east2 >= east1))
                    {
                      a = east2 - e_plane;
                      b = n_plane - north2;
                      c = (a * a) + (b * b);
                      if (c == 0)
                        {
                          xf = east2;
                          yf = north2;
                        }
                      else
                        {
                          d = sqrt(c);
                          x1 = (a / d) * flightdis;
                          y1 = (b / d) * flightdis;
                          x2 = ( b / d) * wingdis;
                          y2 = ( a / d) * wingdis;
                          xf = e_plane + (x1 + x2);
                          yf = n_plane - (y1 - y2);
                        }
                    }
                  /* Southwest Direction */
                  if ((north2 < north1) && (east2 < east1))
                    {
                      a = e_plane - east2;
                      b = n_plane - north2;
                      c = (a * a) + (b * b);
                      if (c == 0)
                        {
                          xf = east2;
                          yf = north2;
                        }
                      else
                        {
                          d = sqrt(c);
                          x1 = (a / d) * flightdis;
                          y1 = (b / d) * flightdis;
                          x2 = ( b / d) * wingdis;
                          y2 = ( a / d) * wingdis;
                          xf = e_plane - (x1 - x2);
                          yf = n_plane - (y1 + y2);
                        }
                    }

                  //point1->acheight = altitude;
                  point1->gndalt = point1->alt - altitude;
                  point1->gndnorth = yf;
                  point1->gndeast = xf;
                }
            skipit:
              if (point1->next == point2) flag = false;
              else point1 = point1->next;
            }
        }
    }
  return;
}

bool CFlightData::QueueLasWat(bool las, double time)
{
  LAB *temp = GetLab();
  if (!temp)
    {
      if (las) Out2User("Memory Error Queuing the Laser");
      else Out2User("Memory Error Queuing the Watson");
      CloseFiles();
      ClearList();
      return false;
    }
  LAB *here = GetLab();
  if (!here)
    {
      delete temp;
      if (las) Out2User("Memory Error Queuing the Laser");
      else Out2User("Memory Error Queuing the Watson");
      CloseFiles();
      ClearList();
      return false;
    }

  do
    {
      ReadStatus status;

      memcpy(here, temp, sizeof(LAB));

      if (las) status = LaserRead(temp);
      else status = WatsonRead(temp);
      if (status != rGOOD)
        {
          delete temp;
          delete here;
          if (las) Out2User("Error scanning Laser File to Match Time with Horita File");
          else Out2User("Error scanning Watson File to Match Time with Horita File");
          CloseFiles();
          ClearList();
          return false;
        }
    }
  while (temp->labtime < time);

  // add to list
  if (here->labtime == 0.0)
    {
      delete here;
      temp->next = m_masterList;
    }
  else
    {
      here->next = m_masterList;
      temp->next = here;
    }
  m_masterList = temp;

  return true;
}

bool CFlightData::LoadLasWat(bool las, double time, ReadStatus *stat)
{
  LAB *temp;

  do
    {
      temp = GetLab();
      if (!temp)
        {
          if (las) Out2User("Memory Error Loading the Laser");
          else Out2User("Memory Error Loading the Watson");
          CloseFiles();
          ClearList();
          return false;
        }

      if (las) *stat = LaserRead(temp);
      else *stat = WatsonRead(temp);
      if (*stat == rFILE_END)
        {
          delete temp;
          return true;
        }
      if (*stat == rBAD_SCAN)
        {
          delete temp;
          if (las) Out2User("Error scanning Laser File to Match Time with Horita File");
          else Out2User("Error scanning Watson File to Match Time with Horita File");
          CloseFiles();
          ClearList();
          return false;
        }

      // add to list
      temp->next = m_masterList;
      m_masterList = temp;
    }
  while(temp->labtime < time);

  return true;
}

bool CFlightData::ComputeGP(double n1, double e1, double n2, double e2, double laser, float roll, float pitch, int frame)
{
  LAB *temp;
  char buf[36];

  // set composer flags and ensure list is cleared
  m_composerFlags[cLASER_ALT] = false;
  m_composerFlags[cLASER_GND] = false;
  m_composerFlags[cLASER_AVE] = false;
  m_composerFlags[cLASER_FILE] = true;
  ClearList();

  // create last entry
  temp = GetLab();
  if (!temp)
    {
      Out2User("Memory Error");
      return false;
    }
  temp->north = n2;
  temp->east = e2;
  temp->labtime = 30.0;
  temp->fHorita = 2;

  // add to list
  m_masterList = temp;

  // create second entry
  temp = GetLab();
  if (!temp)
    {
      Out2User("Memory Error(2)");
      ClearList();
      return false;
    }
  temp->laser = laser;
  temp->alt = 100.0;
  temp->roll = roll;
  temp->pitch = pitch;
  temp->labtime = (double) frame;
  temp->north = ((n2 - n1) * ((double) frame) / 30.0) + n1;
  temp->east = ((e2 - e1) * ((double) frame) / 30.0) + e1;

  // add to list
  temp->next = m_masterList;
  m_masterList = temp;

  // create first entry
  temp = GetLab();
  if (!temp)
    {
      Out2User("Memory Error(3)");
      ClearList();
      return false;
    }
  temp->north = n1;
  temp->east = e1;
  temp->fHorita = 2;

  // add to list
  temp->next = m_masterList;
  m_masterList = temp;

  // compute ground point
  GPFinder();

  temp = m_masterList->next;
  sprintf(buf,"GndNorth = %1.2lf, GndEast = %1.2lf", temp->gndnorth, temp->gndeast);
  Out2User(buf);

  ClearList();
  return true;
}
