// FlightData.h : interface of the CFlightData class
//
/////////////////////////////////////////////////////////////////////////////
#include <stdio.h>


#define r_pie 3.14159265359

typedef enum ReadStatus {rGOOD, rBAD_SCAN, rFILE_END};
typedef enum FileType {NONE, LASER, WATSON, HORITA, HORITA_RR, HORITA_RC, COMPOSER};
typedef enum CompFlags {cLASER_FILE, cLASER_AVE, cLASER_ALT, cLASER_GND, cWATSON_FILE,
                        cOUTPUT_SINGLE, cLASER_ONLY, cNO_ZEROS, cORIG_HEADING};
const short COMPFLAGSIZE = 9;
typedef enum PlotType {PLOT_ALT, PLOT_PLANE, PLOT_GROUND};

typedef struct s1
{
  double labtime;
  double mSecPerByte;
  char data[64];
  int offset;
  struct s1 *next;
}ENTRY;

typedef struct s3
{
  double labtime;		// computer time for synchronization purposes
  double laser;		        // laser distance in feet
  int scatter;			// laser scatter
  char watsValid;		// Watson 'I' character - valid info indicator
  float roll, pitch;		// Watson roll, pitch angle in degrees
  float heading;		// Watson heading angle in degrees
  float oldheading;
  short hour, min, sec, frame;	// GPS timecode
  short u1, u2, u3, u4;		// GPS user bytes
  char theDate[10];		// Horita date field
  bool fLaser, fWatson;		// raw data flags
  char fHorita;
  double alt, oldalt;		// GPS altitude with respect to Sea Level
  double north, east;		// GPS UTM Northing and Easting
  double oldnorth, oldeast;
  char zone[5];			// GPS UTM Zone
  double gndnorth, gndeast;     // computed ground coordinates (UTM) of laser strike
  double gndalt, localt;	// computed ground altitude (UTM) of laser strike
  struct s3 *next;
}LAB;

// define maximum length of entry list
const short MAXSIZE = 1000;

// define # of seconds to store in masterlist
const short MASTERSIZE = 5;

// define an error buffer
const short ERROR_BUFFER = 7;

// define timestamp tolerances (in milliseconds)
const double LASER_TOLERANCE = 2000.0;
const double WATSON_TOLERANCE = 2000.0;

// define a Frame number tolerance for the Horita Cleaner
const short FRAME_TOL = 2;

class CFlightData
{
 public:
  CFlightData(); // Constructor
  ~CFlightData(); // Destructor

// Attributes
private:
  bool LoadLasWat(bool las, double time, ReadStatus *stat);
  bool QueueLasWat(bool las, double time);
  void GPFinder();
  void CorrectHeading();
  void InterWat();
  void InterGPS();
  void OffsetGPS();
  char UTMLetterDesignator(double Lat);
  void LLtoUTM(double Lat, double Long, double *UTMNorthing, double *UTMEasting, char* UTMZone);
  LAB * Merge(LAB* lsta, LAB *lstb);
  LAB * MergeSort(LAB *lsta, long count);
  LAB * CorrectGPS();
  ReadStatus WatsonRead(LAB *temp);
  ReadStatus LaserRead(LAB *temp);
  ReadStatus HoritaRead(LAB *temp);
  LAB * GetLab();
  double CheckSort(FILE *fp);
  void ProcessUserData(int frame, int u1, int u2, int u3, int u4);
  bool CleanRawFile();
  char Num2HexChar(int x);
  ENTRY * ProcessHorita(ENTRY *here, ENTRY **newEntry, int *index, int *num, unsigned char c);
  int LoadRawHorita();
  int LoadRawWatson();
  ENTRY * RemovePeaks(ENTRY *theEnd);
  ENTRY * RemoveSpikes(ENTRY *theEnd);
  void Out2File(ENTRY *lastTime);
  ENTRY * ComputeRate();
  ENTRY * PrevalQueue();
  void CloseFiles();
  int LoadRawLaser();
  ENTRY * GetEntry();
  void ClearList();
  ENTRY *m_listHead;
  FILE *fpLAS, *fpWAT, *fpHOR, *fpOUT;
  int m_fileNumber;
  FileType m_currentID;
  LAB *m_masterList;

public:
  // cleaner variables
  bool m_analysis;
  bool m_original;
  char m_inputLaserFile[256];
  char m_inputHoritaFile[256];
  char m_inputWatsonFile[256];
  char m_outputFileName[256];
  char m_newFile[256];		// a working buffer
  double m_startLab, m_endLab;
  double averageSum;
  double averageTime;

  // composer variables
  bool m_markLines;
  bool m_composerFlags[COMPFLAGSIZE];
  char m_masterDate[10];
  float m_cameraForward;
  float m_cameraSide;
  float m_cameraVertical;
  float m_rollOffset;
  float m_pitchOffset;
  float m_laserValue;

  PlotType pType;

// Implementation
public:
  bool GenPlot();
  bool ComputeGP(double n1, double e1, double n2, double e2, double laser, float roll, float pitch, int frame);
  bool Composer();
  bool ReadRawHorita();
  void Out2User(char *s);
  bool CleanRawHorita();
  bool CleanRawWatson();
  bool CleanRawLaser();
  void ResetDocument();

};
